import termios
import tty

def enter_raw_mode(fd):
    """Enters raw mode for the given file descriptor."""
    old_settings = termios.tcgetattr(fd) #
    tty.setraw(fd) #
    return old_settings


def exit_raw_mode(fd, settings):
    """Restores the terminal to its original settings."""
    termios.tcsetattr(fd, termios.TCSADRAIN, settings) #


def switch_echo(fd, enabled):
    (iflag, oflag, cflag, lflag, ispeed, ospeed, cc) \
        = termios.tcgetattr(fd)

    if enabled:
        lflag |= termios.ECHO
    else:
        lflag &= ~termios.ECHO

    new_attr = [iflag, oflag, cflag, lflag, ispeed, ospeed, cc]
    termios.tcsetattr(fd, termios.TCSANOW, new_attr)



import os
import sys

def alt_buffer_on(fd):
    pass


def alt_buffer_off(fd):
    pass


def draw_data(data: bytes, dest_fd: int | None = None) -> None:
    if dest_fd is None:
        dest_fd = sys.stdout.fileno()
    
    try:
        os.write(dest_fd, data)
        # sys.stdout.flush()
    except Exception as e:
        print(f"An error occurred: {e}")


from collections.abc import Callable, Awaitable
import termios
import os
import select
import sys
from typing import Self

from pyntercept.pseudo_tty import create_pty

class PTYProcess:
    # width: int
    # height: int
    
    def __init__(
        self, 
        argv: list[str], 
        
        width: int | None = None, height: int | None = None,
        
        in_fd: int | None = None, out_fd: int | None = None, 
        err_fd: int | None = None,
        
        in_upd_callback: Callable[ [Self], bytes | None] | None = None,
        out_upd_callback: Callable[ [Self], bytes | None] | None = None,
    ):
        pid, master_fd, child_fd = create_pty(argv)
        
        self.pid = pid
        self.master_fd = master_fd
        self.child_fd = child_fd
        
        self.in_fd  = (in_fd or sys.stdin.fileno())
        self.out_fd = (out_fd or sys.stdout.fileno())
        self.err_fd = (err_fd or sys.stderr.fileno())
        
        self.in_upd_callback = in_upd_callback
        self.out_upd_callback = out_upd_callback
        
        if width is None or height is None:
            rows, cols = termios.tcgetwinsize(sys.stdout.fileno())
            
            width = width or cols
            height = height or rows
        
        self.set_size(width, height)
    
    
    def set_size(self, width = 60, height = 20):
        termios.tcsetwinsize(self.child_fd, (height, width))
    
    
    def get_size(self) -> tuple[int, int]:
        return termios.tcgetwinsize(self.child_fd)
    
    
    def child_alive(self) -> bool:
        # Availability: Unix, Windows, not WASI, not Android, not iOS.
        rpid, _ = os.waitpid(self.pid, os.WNOHANG)
        return rpid == 0
    
    
    def update(self) -> bool:
        # returns False if process terminated
        
        if not self.child_alive():
            return False
            
        # aks OS for file descriptors updates
        rlist, _, _ = select.select([self.in_fd, self.master_fd], [], [], 0)
        
        return_status = True
        if self.in_fd in rlist:
            if self.in_upd_callback:
                res = self.in_upd_callback(self)
            else:    
                res = self.on_in_fd_upd()
            return_status *= res is not None
            
        if self.master_fd in rlist:
            if self.out_upd_callback:
                res = self.out_upd_callback(self)
            else:
                res = self.on_out_fd_upd()
            return_status *= res is not None
            
        return return_status
    
    
    def on_in_fd_upd(process) -> bytes:
        data =  os.read(process.in_fd, 2048)    # read user input
        os.write(process.master_fd, data)       # pass input into editor process
        
        return data
    
    
    def on_out_fd_upd(process) -> bytes:
        return os.read(process.master_fd, 2048) # read output from editor


import pty
import os
import errno


# code is mainly taken away from ptyprocess
# 

def pty_make_controlling_tty(tty_fd):
    child_name = os.ttyname(tty_fd)
        
    # Disconnect from controlling tty, if any.  Raises OSError of ENXIO
    # if there was no controlling tty to begin with, such as when
    # executed by a cron(1) job.
    try:
        fd = os.open("/dev/tty", os.O_RDWR | os.O_NOCTTY)
        os.close(fd)
    except OSError as err:
        if err.errno != errno.ENXIO:
            raise
            
    os.setsid()

    # Verify we are disconnected from controlling tty by attempting to open
    # it again.  We expect that OSError of ENXIO should always be raised.
    try:
        fd = os.open("/dev/tty", os.O_RDWR | os.O_NOCTTY)
        os.close(fd)
        raise RuntimeError("OSError of errno.ENXIO should be raised.")
    except OSError as err:
        if err.errno != errno.ENXIO:
            raise
    
    # Verify we can open child pty.
    fd = os.open(child_name, os.O_RDWR)
    os.close(fd)

    # Verify we now have a controlling tty.
    fd = os.open("/dev/tty", os.O_WRONLY)
    os.close(fd)


def create_pty(argv: list[str]) -> tuple[int, int, int]:
    parent_fd, child_fd = os.openpty()
    
    if parent_fd < 0 or child_fd < 0:
        raise OSError("os.openpty() failed")
    
    pid = os.fork()
    if pid == pty.CHILD:
        os.close(parent_fd)
        pty_make_controlling_tty(child_fd)
        
        os.dup2(child_fd, pty.STDIN_FILENO)
        os.dup2(child_fd, pty.STDOUT_FILENO)
        os.dup2(child_fd, pty.STDERR_FILENO)
        
        os.execlp(argv[0], *argv)
    # we keep child_fd opening to be able to resize it
    # else: 
    #     os.close(child_fd)
        
    return pid, parent_fd, child_fd




#!/usr/bin/env python3

import sys

from pyntercept.draw import draw_data
from pyntercept.tty_utils import enter_raw_mode, exit_raw_mode, switch_echo
from pyntercept.process import PTYProcess

stdin_fd = sys.stdin.fileno()
stdout_fd = sys.stdin.fileno()

def on_out_upd(process: PTYProcess) -> bytes:
    data = process.on_out_fd_upd()

    draw_data(data, stdout_fd)
    
    return data

def main():    
    if len(sys.argv) < 2:
        print('you must specify executable program!')
        return
    
    pty_process = PTYProcess(
        sys.argv[1:],
        out_upd_callback=on_out_upd
    )
    
    try:
        switch_echo(stdin_fd, False)
        old_stdin = enter_raw_mode(stdin_fd)
        old_stdout = enter_raw_mode(stdout_fd)
        
        draw_data(pty_process.on_out_fd_upd())
        
        while pty_process.update():
            pass
        
    except OSError: # this happens when child process dies
        pass
    finally:
        exit_raw_mode(stdin_fd, old_stdin)
        exit_raw_mode(stdout_fd, old_stdout)
        switch_echo(stdin_fd, True)


if __name__ == "__main__":
    main() 